function [L, D, tIQ, args] = additive_asymmetric_step(s, gamma, h, B, IQL, IQD, L0, D0, Rinv, args, t0)
% Do a step of size h with the s-order additive asymmetric splitting method 
%
%       gamma(1) exp(hF)exp(hG)P0 )
%       gamma(2) (exp(h/2F)exp(h/2G))^2 P0 + 
%         ... +
%       gamma(s) (exp(h/sF)exp(h/sG))^s P0 
% where
%       P0 = L0*D0*L0'
%
% Here F(P) = M^{-T} A^T P + PAM^{-1} + M^{-T}QM^{-1} and
% G(P) = -P B Rinv B^T P.

% Further, IQL{j}*IQD{j}*IQL{j}^T should be a low-rank factorization of
% \int_{0}^{h/j}{ exp(s(A M^{-1})^T)M^{-T} Q M^{-1} exp(s A M^{-1}) ds}

%
% Copyright 2018 Tony Stillfjord
%
% This file is part of DREsplit.
% 
% DREsplit is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% DREsplit is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with DREsplit.  If not, see <http://www.gnu.org/licenses/>.
%

    outkeys = {}; % For storing new temporary data
    outdata = {};
    keys = args.impODEobj.keys;
    tIQj = {};
    
    if ~args.timevarying
        parfor j = 1:s
            L2{j} = L0; D2{j} = D0;

            for k = 1:j
                [L2{j}, D2{j}] = expF(h / j, IQL{j}, IQD{j}, L2{j}, D2{j}, args);
                [L2{j}, D2{j}] = expG(h / j, B, Rinv, L2{j}, D2{j}, args);
            end
            [outkeys{j}, ind] = setdiff(args.impODEobj.keys, keys);
            outdata{j} = args.impODEobj.data(ind);
        end
        tIQ = 0;
    else
        LQ = args.LQ;
        DQ = args.DQ;
        parfor j = 1:s
            L2{j} = L0; D2{j} = D0;
            
            % Approximate the integral term over the relevant intervals
            tIQj = tic;
            [IQL, IQD] = compute_integral_factors_LDLT(LQ, DQ, h, 1/j, args, t0 + h*(0:j-1)/j);
            tIQ{j} = toc(tIQj);

            for k = 1:j
                t = t0 + (k-1) / j * h;
                [L2{j}, D2{j}] = expG(h / j, B, Rinv, L2{j}, D2{j}, args, t);
                [L2{j}, D2{j}] = expF(h / j, IQL{k}{1}, IQD{k}{1}, L2{j}, D2{j}, args, t);
            end
            [outkeys{j}, ind] = setdiff(args.impODEobj.keys, keys);
            outdata{j} = args.impODEobj.data(ind);
        end
        tIQ = sum(cell2mat(tIQ));
    end
    
    % Add new temporary data
    newkeys = [];
    newdata = {};
    for j = 1:s
        % If the parfor is actually a for, args will already have been
        % updated, so check again:
        outkeys{j} = setdiff(outkeys{j}, args.impODEobj.keys);
        
        for k = 1:length(outkeys{j})
            if ~any(newkeys == outkeys{j}(k)) % This is really a new key
                newkeys(end+1) = outkeys{j}(k);
                newdata{end+1} = outdata{j}{k};
            end
        end
    end
    
    args.impODEobj.keys = [args.impODEobj.keys, newkeys];
    args.impODEobj.data = [args.impODEobj.data, newdata];

    % Stack the results
    L = cell2mat(L2);

    for j = 1:s
        D2{j} = gamma(j)*D2{j};
    end
    D = blkdiag(D2{:});

    [L, D] = column_compress_LDLT(L, D, args.CC_tol);

end